#!/usr/bin/env python3

# -*- coding: utf-8 -*-

import importlib
import inspect
import sys

# mixin for ansible.plugins.filters.*

class FilterModule(object):

  def __init__(self):
    self._filters = {}

  def filters(self):

    for spec in [
        'ansible.plugins.filter.core',
        'ansible.plugins.filter.mathstuff',
        'ansible.plugins.filter.urls',
        'ansible.plugins.filter.urlsplit',
      ]:

      mod = importlib.import_module(spec)

      if hasattr(mod, 'FilterModule'):
        _filters = mod.FilterModule().filters()
        for k,v in _filters.items():
          self._filters.update( {
            k: [ "{}.{}".format(spec, k), v ]
          } )

      del mod
    return self._filters

class TestModule(object):

  def __init__(self):
    self._tests = {}

  def tests(self):

    for spec in [
        'ansible.plugins.test.core',
        'ansible.plugins.test.files',
        'ansible.plugins.test.mathstuff',
      ]:

      mod = importlib.import_module(spec)

      if hasattr(mod, 'TestModule'):
        _tests = mod.TestModule().tests()
        for k,v in _tests.items():
          self._tests.update( {
            k: [ "{}.{}".format(spec, k), v ]
          } )

      del mod
    return self._tests



#!/usr/bin/env python3

# -*- coding: utf-8 -*-

# https://jinja.palletsprojects.com/en/2.11.x/templates/#list-of-builtin-filters
# https://www.webforefront.com/django/usebuiltinjinjafilters.html

import re
import sys
import os
import csv

# TODO
# Encapsulating filters in this way is great but it does little for their
# test coverage (nice trick to fool pycoverage eh!)
# Set these up for proper unit (not end-to-end) tests.

filters = dict(

  env_override = (
  """
    Allow for a value set in the environment to override a given value
    e.g. url | env_override("URL")
    """,
    lambda v, k: os.getenv(k, v)
  ),

  format_dict = (
  """
    Given a dict as value, return a formatted string as specified in format.
    e.g. url | urlsplit | format_dict("{scheme}://{hostname}:8080/{path}/")
    """,
    lambda v, f: f.format(**v)
  ),

  format_list = (
  """
    Given a list as value, return a formatted string as specified in format.
    e.g. url | urlsplit | format_list("{0}://{1}")
    """,
    lambda v, f: f.format(*v)
  ),

  items = ( """
    Select items specified by indexes from the list passed in
    e.g. range(1,10) | items(0,2,5,-3,-2)
    """,
    lambda *n: list(n[0][x] for x in n[1:])
  ),

  keys = (
    """
    Return the keys of a dict passed in
    """,
    lambda d: d.keys()
  ),

  values = (
    """
    Return the values of a dict passed in
    """,
    lambda d: d.values()
  ),

  wrap = ( """
    Wrap value with "parantheses" specified in format (default "()")
    """,
    lambda v, t='()': '{}{}{}'.format(t[0], v, t[1])
  ),

  to_csv = ( """
    Split a string delimited by commas and return items in a list
    """,
    lambda v: list(csv.reader(v))[0]  # magic number is for single line of text
  ),

  to_url = ( """
    Take values from a dict passed in and
    return a string formatted in the form of a URL
    """,
    lambda v, f='https://{hostname}':
                  f.format(**v) ),
)

if 'USE_ANSIBLE_SUPPORT' in os.environ.keys():
  from .ansible import FilterModule
  filters.update(FilterModule().filters())

for k,v in filters.items():
  setattr(sys.modules[__name__], k, v[1])



#!/usr/bin/env python3

# -*- coding: utf-8 -*-

# Hold custom jinja2 tests
# https://jinja.palletsprojects.com/en/2.11.x/api/#custom-tests
# https://jinja.palletsprojects.com/en/2.11.x/templates/#list-of-builtin-tests

from jinja2 import is_undefined
import re
import sys
import math
import os

def _is_prime(n):
  if n == 2:
    return True
  for i in range(2, int(math.ceil(math.sqrt(n))) + 1):
    if n % i == 0:
      return False
  return True

tests = dict(

  is_prime = ( """ Tests if a number is prime """,
    lambda v: _is_prime(v)
  ),

)

if 'USE_ANSIBLE_SUPPORT' in os.environ.keys():
  from .ansible import TestModule
  tests.update(TestModule().tests())

for k,v in tests.items():
  setattr(sys.modules[__name__], k, v[1])



#!/usr/bin/env python3

# -*- coding: utf-8 -*-

# NAME
# inji/__init__.py

# Copyright (C) 2020 Shalom Bhooshi
# Author: Shalom Bhooshi

""" Utility for rendering jinja2 templates """

from . import cli

__version__  = cli._version()
cli_location = cli.cli_location()



#!/usr/bin/env python3

# -*- coding: utf-8 -*-

import argparse
from os.path import abspath, basename, dirname, exists, expandvars, isdir, isfile, join
import json
import os
import fnmatch
import sys
import yaml

def json_parse(string):
  """ Parse a JSON string into a dictionary """
  try:
    return json.loads(string)
  except Exception as e:
    msg = 'Error parsing JSON config: {}'.format(str(e))
    print(msg, file=sys.stderr)
    raise TypeError(msg)


def kv_parse(string):
  """ Parse a string of the form foo=bar into a dictionary """
  try:
    key, val = string.split('=', 1)
    if key is None or key == '':
      raise TypeError('Empty key')
  except Exception as e:
    err = "Invalid key found parsing KV string '{}': {}".format(string, str(e))
    print(err, file=sys.stderr)
    raise
  return { key: val }


def read_context(yaml_file):
  yaml_file = yaml_file.__str__()
  with open(yaml_file, 'r') as f:
    try:
      in_vars = yaml.load(f, Loader=yaml.SafeLoader)
      if in_vars is None:
        raise TypeError("'{}' contains no data".format(yaml_file))
    except TypeError as exc:
      raise exc
  return in_vars


def recursive_iglob(rootdir='.', pattern='*'):
  for root, dirnames, filenames in os.walk(rootdir):
    for filename in fnmatch.filter(filenames, pattern):
      yield os.path.join(root, filename)


def path(fspath, type='file'):
  """
  Checks if a filesystem path exists with the correct type
  """

  fspath = abspath(expandvars(str(fspath)))
  msg = None
  prefix = "path '{0}'".format(fspath)

  if not exists(fspath):
    msg = "{0} does not exist".format(prefix)

  if type == 'file' and isdir(fspath):
    msg = "{0} is not a file".format(prefix)

  if msg is not None:
    print(msg, file=sys.stderr)
    raise argparse.ArgumentTypeError(msg)

  return fspath


def file_or_stdin(file):
  # /dev/stdin is a special case allowing bash (and other shells?) to name stdin
  # as a file. While python should have no problem reading from it, we actually
  # read template relative to the template's basedir and /dev has no templates.
  if file == '-' or file == '/dev/stdin':
    return '-'
  return path(file)



#!/usr/bin/env python3

# -*- coding: utf-8 -*-

from jinja2 import DebugUndefined, StrictUndefined, Undefined, make_logging_undefined
from jinja2 import Environment, FileSystemLoader
from jinja2.exceptions import TemplateNotFound, UndefinedError
import sys
import logging
import inspect

from . import utils
from . import filters
from . import tests
from . import globals

def get_symbols(mod):
  return { k:v for k,v
            in inspect.getmembers(mod)
            if not (k.startswith('_')) }

class TemplateEngine(object):

  def __init__( self,
                undefined_variables_mode_behaviour='strict',
                j2_env_params={},
    ):

    UndefinedHandler = StrictUndefined
    m = undefined_variables_mode_behaviour
    if   m in ['empty',  'Undefined']:
      UndefinedHandler = Undefined
    elif m in ['keep',   'DebugUndefined']:
      UndefinedHandler = DebugUndefined

    # Setup debug logging on STDERR to have the jinja2 engine emit
    # its activities
    root = logging.getLogger(__name__)
    root.setLevel(logging.DEBUG)
    handler = logging.StreamHandler(sys.stderr)
    handler.setFormatter(logging.Formatter('%(name)s %(levelname)s: %(message)s'))
    root.addHandler(handler)

    UndefinedHandler = make_logging_undefined( logger=root, base=UndefinedHandler )

    j2_env_params.setdefault('undefined', UndefinedHandler)
    j2_env_params.setdefault('trim_blocks', True)
    j2_env_params.setdefault('keep_trailing_newline', False)
    j2_env_params.setdefault('extensions', [
                            'jinja2.ext.i18n',
                            'jinja2.ext.do',
                            'jinja2.ext.loopcontrols',
                          ])

    self.j2_env_params = j2_env_params

    self.filters = get_symbols(filters)
    self.tests   = get_symbols(tests)
    self.globals = get_symbols(globals)

  def render(
      self,
      template,
      context
    ):
    """ Render the template """

    # We don't assume that includes and other sourceables reside relative
    # to the current directory but instead relative to the "master" template
    # we are processing. We deviate from jinja tradition this way.
    # This is because.
    # 1. We process template from stdin in temporary directories
    # 2. We should be free to change the process CWD and not break rendering
    rootdir = utils.dirname(template)

    self.j2_env_params.setdefault('loader', FileSystemLoader(rootdir))
    j2_env = Environment(**self.j2_env_params)

    j2_env.globals.update(self.globals)
    j2_env.filters.update(self.filters)
    j2_env.tests.update(self.tests)

    try:
      template = utils.basename(template)
      yield j2_env.get_template(template).render(context)
    except UndefinedError as e:
      raise UndefinedError( "variable {} in template '{}'".format(
              str(e), template) ) from e




#!/usr/bin/env python3

# -*- coding: utf-8 -*-

# hold global variables and functions for direct use in templates
# https://jinja.palletsprojects.com/en/2.11.x/api/#the-global-namespace
# https://jinja.palletsprojects.com/en/2.11.x/templates/#list-of-global-functions

from datetime import datetime, timezone
import socket
import sys

_globals = dict(

  now  = ( """ Return the timestamp for datetime.now() """,
    lambda : datetime.now()
  ),

  date  = ( """ Return the timestamp for datetime.now() """,
    datetime.now()  # variable
  ),

  strftime = ( """ Return a date string specified in strftime(3) format """,
    lambda f='%F %T%z', tz='UTC':
      datetime.now(timezone.utc if tz == 'UTC' else None).strftime(f)
  ),

  hostname = ( """ Return the current hostname """,
    lambda : socket.gethostname()
  ),

)

for k,v in _globals.items():
  setattr(sys.modules[__name__], k, v[1])



#!/usr/bin/env python3

# -*- coding: utf-8 -*-

import distutils.cmd
import os
import re
import setuptools
import subprocess
import sys

# Suppress setup.py's stdout logging for our commands that output data
if any( x in ['requirements', 'version'] for x in sys.argv ):
    sys.argv.insert(1, '-q')

requirements_dev = '''
jinja2==2.11.*
pyyaml==5.3.*
'''.strip().split('\n')

requirements_test = '''
ansible>=2.8
cffi==1.14.2
coverage==5.2.1
pytest-tap==3.1
pytest-cov==2.10.1
pytest==6.0.1
'''.strip().split('\n')

# https://jichu4n.com/posts/how-to-add-custom-build-steps-and-commands-to-setuppy/
class RequirementsCommand(distutils.cmd.Command):
    """ Emit requirements """
    description = 'emit requirements'
    user_options = [
        ('dev',  None, 'emit dev requirements'),
        ('test', None, 'emit test requirements')
    ]

    def initialize_options(self):
        self.dev  = None
        self.test = None

    def finalize_options(self): pass

    def run(self):
        if self.test:
           requirements = requirements_test
        else:
           requirements = requirements_dev
        print('\n'.join(requirements))
        return requirements

def version():
    args = 'git describe --tags --always'.split(' ')
    version = subprocess.check_output(args).decode().strip()
    version = re.sub('^[vV]|\-\w{8}$', '', version)
    version = re.sub('-', '.post', version)
    return version

class VersionCommand(distutils.cmd.Command):
    """ Emit version """
    description = 'Emit version'
    user_options = [
        ('next', None, 'next'),
        ('version', None, 'version')
    ]
    def initialize_options(self): pass
    def finalize_options(self): pass
    def run(self):
        print(version())
        return(version())

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
        name             = 'inji',
        version          = version(),
        description      = 'Render parametrized Jinja2 templates at the CLI',
        author           = 'Shalom Bhooshi',
        author_email     = 's.bhooshi@gmail.com',
        license          = 'Apache License 2.0',
        url              = 'https://github.com/shalomb/inji',
        download_url     = 'https://github.com/shalomb/inji/tarball/{}'.format(version()),
        packages         = setuptools.find_packages(),
        scripts          = [ 'bin/inji' ],
        install_requires = requirements_dev,
        include_package_data = True,
        zip_safe         = False,
        python_requires  = '>=3.5',
        long_description_content_type = 'text/markdown',
        long_description = long_description,
        keywords         = [ 'jinja', 'jinja2', 'templating' ],
        classifiers      = [
            'Development Status :: 4 - Beta',
            'Environment :: Console',
            'Intended Audience :: Developers',
            'Intended Audience :: System Administrators',
            'License :: OSI Approved :: Apache Software License',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 3',
            'Topic :: Software Development',
            'Topic :: System :: Systems Administration'
        ],
        cmdclass         = {
            'requirements': RequirementsCommand,
            'version':      VersionCommand,
        }
    )



#!/usr/bin/env python3

# -*- coding: utf-8 -*-

# NAME

# inji - Render jina2 templates to stdout

import argparse
import atexit
import fnmatch
import json
import locale
import os
from   os.path import abspath, dirname, join
import pkg_resources
import re
import shutil
import signal
import sys
import tempfile

from .engine import TemplateEngine
from . import utils

def pkg_location():
  return abspath(dirname(__file__))

def cli_location():
  return abspath(join(pkg_location(), '../bin/inji'))

def _version():
  return pkg_resources.require(__package__)[0].version

__version__ = _version()

def cli_args():
  parser = argparse.ArgumentParser(
      description='inji - render jinja templates'
    )
  required = parser.add_argument_group('required arguments')

  parser.add_argument('-j', '--json-config', '-c',
    action = 'store', required=False,
    type=lambda x: utils.json_parse(x),
    dest='json_string',
    help="-c '{ \"foo\": \"bar\", \"fred\": \"wilma\" }'"
  )

  parser.add_argument('-k', '--kv-config', '-d',
    action = 'append', required=False,
    type=lambda x: utils.kv_parse(x),
    dest='kv_pair',
    help='-d foo=bar -d fred=wilma'
  )

  parser.add_argument('-o', '--overlay-dir',
    action = 'append', required=False, type=lambda p, t='dir': utils.path(p, t),
    dest='overlay_dir', default=[],
    help='/path/to/overlay/'
  )

  parser.add_argument('-v', '-p', '--vars-file', '--vars',
    action = 'append', required=False, type=lambda p, t='file': utils.path(p, t),
    dest='vars_file', default=[],
    help='/path/to/vars.yaml'
  )

  parser.add_argument('--strict-mode', '-s',
    action = 'store', required=False, type=str,
    dest='undefined_variables_mode', default='strict',
    choices=[ 'strict', 'empty', 'keep',
              'StrictUndefined', 'Undefined', 'DebugUndefined' ],
    help='Refer to http://jinja.pocoo.org/docs/2.10/api/#undefined-types'
  )

  parser.add_argument('--version',
    action = 'version',
    version=_version(),
    help='print version number ({})'.format(_version())
  )

  parser.add_argument('template',
    nargs='*',
    action = 'store',
    type=utils.file_or_stdin,
    default='-',
    help='/path/to/template.j2 (defaults to -)'
  )

  return parser.parse_args()


def sigint_handler(signum, frame):  # pragma: no cover # despite being covered
  """ Handle SIGINT, ctrl-c gracefully """
  signal.signal(signum, signal.SIG_IGN) # ignore subsequent ctrl-c's
  sys.exit( 128 + signal.SIGINT )       # 130 by convention


def main():
  """ Our main method """

  # cleanly handle ctrl-c's
  signal.signal(signal.SIGINT, sigint_handler)

  assert sys.version_info >= (3,5), 'Python version ({}.{} !>= 3.5)'.format(
    sys.version_info.major,
    sys.version_info.minor
  )

  args = cli_args()

  # this holds all the possible vars files we are told about or imply
  vars_files = []

  # context in the local configuration files - p5
  vars_files += fnmatch.filter(os.listdir('.'), "*inji.y*ml")

  # context in the overlay directories - p4
  for d in args.overlay_dir:
    files = list(utils.recursive_iglob(d, '*.y*ml'))
    if len(files):
      loc = locale.getlocale()
      locale.setlocale(locale.LC_ALL, 'C') # Using LC_ALL=C POSIX locale
      files.sort()                         # We force the sort collation of files
      locale.setlocale(locale.LC_ALL, loc) # And then reset it back
      vars_files += files

  # context from named vars files - p3
  vars_files += args.vars_file

  # This will hold the final vars dict merged from various available sources
  context = {}
  for file in vars_files:
    context.update(utils.read_context(file))

  # context from environment variables - p2
  context.update(os.environ)

  # context at the command line (either JSON or KV type) - p1
  if args.json_string:
    context.update(args.json_string)

  if args.kv_pair:
    # we've appended dicts into args.kv_pair .. unpack them in order
    for d in args.kv_pair:
      context.update(d)

  if '-' in args.template:
    # Template passed in via stdin. Create template as a tempfile and use it
    # instead but since includes are possible (though not likely), we have to do
    # this in an isolated tmpdir container to prevent inadvertent reading of
    # includes not meant to be read.
    tmpdir = tempfile.mkdtemp(prefix=__name__)
    atexit.register(shutil.rmtree, tmpdir)

    _, tmpfile = tempfile.mkstemp(prefix='stdin-', dir=tmpdir, text=True)
    atexit.register(os.remove, tmpfile)

    with open(tmpfile, "a+") as f:
      f.write(sys.stdin.read())

    # Yes, even if user specifies multiple other templates, the fact he
    # specified '-' just once means we only deal with one template i.e. '-'
    args.template = [tmpfile]

  engine = TemplateEngine( undefined_variables_mode_behaviour=args.undefined_variables_mode )

  for template in args.template:
    for block in engine.render( template=template, context=context ):
      print(block)
